@@ -1,8 +1,14 @@
-import pika, requests, threading, json, time
+from concurrent.futures.thread import ThreadPoolExecutor
+
+import json
+import pika
+import requests
+import time
 from logzero import logger
 from pymitter import EventEmitter
-from autosolveclient.autosolveconstants import AutoSolveConstants
+
 from autosolveclient.autosolve_errors import AutoSolveConnectionError
+from autosolveclient.autosolveconstants import AutoSolveConstants
 
 
 class Instance:
@@ -34,6 +40,7 @@ class InternalAutoSolve:
         self.client_key = params["client_key"]
         self.debug = params["debug"]
         self.should_alert_on_cancel = params["should_alert_on_cancel"]
+        self.executor = ThreadPoolExecutor(max_workers=10)
         self.queue = []
 
         self.routing_key = ""
@@ -48,11 +55,10 @@ class InternalAutoSolve:
         self.cancel_send_routing_key = ""
 
         self.delay = None
-        self.attempts = None
+        self.attempts = 0
         self.connection = None
         self.channel = None
         self.cancel_channel = None
-        self.consumer_thread = None
         self.ready = False
         self.connected = False
 
@@ -65,12 +71,10 @@ class InternalAutoSolve:
         return self.ready
 
     def init(self, access_token=None, api_key=None):
-        self.attempts = 0
         self.delay = self.autosolve_constants.RETRY_DELAY
         self.connection = None
         self.channel = None
         self.cancel_channel = None
-        self.consumer_thread = None
         self.ready = False
         self.connected = False
 
@@ -103,13 +107,19 @@ class InternalAutoSolve:
         self.api_key = api_key
 
     def begin_connection(self):
+        self.executor.submit(self.process_begin_connection)
+
+    def process_begin_connection(self):
+        sleep_time = self.get_delay(self.delay)
+        self.delay += 1
+        if sleep_time > 0:
+            time.sleep(sleep_time)
         if self.check_input_values():
             self.close_connection()
             status_code = self.validate_connection()
             if status_code == 200:
                 self.log("Validation complete", self.autosolve_constants.SUCCESS)
-                self.consumer_thread = threading.Thread(target=self.connect)
-                self.consumer_thread.start()
+                self.connect()
             elif status_code == 400:
                 self.log(self.autosolve_constants.INVALID_CLIENT_KEY, self.autosolve_constants.ERROR)
                 raise AuthException(self.autosolve_constants.INVALID_CLIENT_KEY)
@@ -124,10 +134,6 @@ class InternalAutoSolve:
 
     def connect(self):
         self.log("Beginning connection establishment", self.autosolve_constants.STATUS)
-        sleep_time = self.get_delay(self.delay)
-        self.delay += 1
-        if sleep_time > 0:
-            time.sleep(sleep_time)
 
         credentials = pika.PlainCredentials(self.account_id, self.access_token)
         parameters = pika.ConnectionParameters(host=self.autosolve_constants.HOSTNAME,
@@ -136,61 +142,45 @@ class InternalAutoSolve:
                                                credentials=credentials,
                                                connection_attempts=10,
                                                retry_delay=5,
-                                               blocked_connection_timeout=60)
+                                               blocked_connection_timeout=10,
+                                               heartbeat=10)
+
         try:
-            self.connected = self.establish_connection(parameters) \
-                             and self.create_channels() \
-                             and self.register_response_queue()
-
-            if self.connected:
-
-                self.log("Beginning message consumption", self.autosolve_constants.SUCCESS)
-                self.delay = 0
-                self.ready = True
-                self.channel.start_consuming()
-                self.log("Stopped Message Consuming", self.autosolve_constants.SUCCESS)
-                self.emitter.emit(self.autosolve_constants.ERROR_EVENT, AutoSolveConnectionError.CONNECTION_CLOSED_UNEXPECTEDLY)
-            else:
-                self.log("Error with RabbitMQ connection, attempting to re-establish", self.autosolve_constants.WARNING)
-                self.handle_connection_error()
-
-        except pika.exceptions.AMQPConnectionError or pika.exceptions.ConnectionClosedByBroker:
-            self.log("Error with RabbitMQ connection, attempting to re-establish", self.autosolve_constants.WARNING)
-            self.emitter.emit(self.autosolve_constants.ERROR_EVENT, AutoSolveConnectionError.CONNECTION_ERROR_INIT)
+            self.connection = self.establish_connection(parameters)
+            self.connection.ioloop.start()
+        except Exception as e:
+            self.log(e, self.autosolve_constants.ERROR)
             self.handle_connection_error()
 
-        except Exception:
-            self.log("Unknown error occurred during connection.", self.autosolve_constants.ERROR)
-            self.emitter.emit(self.autosolve_constants.ERROR_EVENT, AutoSolveConnectionError.CONNECTION_ERROR_INIT)
-            self.handle_connection_error()
+    def handle_success(self):
+        self.log("Beginning message consumption", self.autosolve_constants.SUCCESS)
+        self.delay = 0
+        self.ready = True
+        self.executor.submit(self.attempt_message_backlog_send)
 
     def close_connection(self):
         self.log("Closing Connection", self.autosolve_constants.STATUS)
         if self.connection is not None and self.connection.is_open:
             self.connection.close()
+            self.connection.ioloop.stop()
         else:
             self.log("Connection does not exist", self.autosolve_constants.STATUS)
 
     def establish_connection(self, parameters):
         self.log("Attempting connection establishment", self.autosolve_constants.STATUS)
+        return pika.SelectConnection(parameters=parameters,
+                                         on_open_callback=self.create_channels,
+                                         on_close_callback=self.handle_connection_error)
 
-        try:
-            self.connection = pika.BlockingConnection(parameters)
-        except Exception:
-            self.log("Connection Exception", self.autosolve_constants.ERROR)
-
+    def create_channels(self, connection):
         self.log("Connection established", self.autosolve_constants.SUCCESS)
-        return True
-
-    def create_channels(self):
         self.log("Creating channel", self.autosolve_constants.STATUS)
-        self.channel = self.connection.channel()
+        self.channel = self.connection.channel(on_open_callback=self.register_response_queue)
         self.cancel_channel = self.connection.channel()
-
         self.log("Channels created", self.autosolve_constants.SUCCESS)
         return True
 
-    def register_response_queue(self):
+    def register_response_queue(self, channel):
         try:
             self.log("Binding queue to exchange", self.autosolve_constants.STATUS)
             self.channel.queue_bind(queue=self.receiver_queue_name,
@@ -203,7 +193,7 @@ class InternalAutoSolve:
                      self.autosolve_constants.SUCCESS)
             self.channel.basic_consume(queue=self.receiver_queue_name, auto_ack=self.autosolve_constants.AUTO_ACK,
                                        on_message_callback=self.receive_message)
-            return True
+            self.handle_success()
         except pika.exceptions.ChannelClosedByBroker:
             self.emitter.emit(self.autosolve_constants.ERROR_EVENT, AutoSolveConnectionError.CONNECTION_ERROR_INIT)
 
@@ -223,7 +213,7 @@ class InternalAutoSolve:
         self.log("Message Received: " + json_string, self.autosolve_constants.SUCCESS)
         self.emitter.emit(self.autosolve_constants.CANCEL_RESPONSE_EVENT, json_string)
 
-    def handle_connection_error(self):
+    def handle_connection_error(self, connection, exception):
         self.connected = False
         self.attempts += 1
 
@@ -233,15 +223,61 @@ class InternalAutoSolve:
         except Exception as ex:
             print(ex)
 
-        if self.connected:
-            self.log("Connection attempt successful, sending any unsent messages", self.autosolve_constants.SUCCESS)
-            self.attempt_message_backlog_send()
-        else:
-            self.log("Connection attempts unsuccessful", self.autosolve_constants.ERROR)
-            self.emitter.emit(self.autosolve_constants.ERROR_EVENT, AutoSolveConnectionError.CONNECTION_REESTABLISH_FAILED)
+    def send_token_request(self, message):
+        self.executor.submit(self.process_send_token_request, message)
+
+    def cancel_token_request(self, taskId):
+        self.executor.submit(self.process_cancel_token_request, taskId)
+
+    def cancel_all_requests(self):
+        self.executor.submit(self.process_cancel_all_requests)
+
+    def process_send_token_request(self, message):
+        self.log("Sending request message for task: " + str(message['taskId']), self.autosolve_constants.STATUS)
+        try:
+            self.queue.append(
+                {"message": message, "route": self.token_send_routing_key, "exchange": self.directExchangeName})
+            send_result = self.send()
+            if send_result is None:
+                self.log("Message with TaskId: " + str(message['taskId']) + " sent successfully",
+                         self.autosolve_constants.SUCCESS)
+        except Exception as e:
+            self.log(e, self.autosolve_constants.WARNING)
+            self.executor.submit(self.resend, message, self.token_send_routing_key, self.directExchangeName)
+
+    def process_cancel_token_request(self, taskId):
+        message = dict()
+        message["taskId"] = taskId
+        message["responseRequired"] = self.should_alert_on_cancel
+
+        self.log("Sending cancel message for task: " + str(message['taskId']), self.autosolve_constants.STATUS)
+        try:
+            self.queue.append({"message": message, "route": self.cancel_send_routing_key,
+                               "exchange": self.fanoutExchangeName})
+            send_result = self.send()
+            if send_result is None:
+                self.log("Cancel message with TaskId: " + str(message['taskId']) + " sent successfully",
+                         self.autosolve_constants.SUCCESS)
+        except Exception as e:
+            self.log(e, self.autosolve_constants.WARNING)
+            self.executor.submit(self.resend, message, self.cancel_send_routing_key, self.fanoutExchangeName)
+
+    def process_cancel_all_requests(self):
+        message = {"apiKey": self.api_key}
+        self.log("Sending cancel all request for api key ::  " + str(message['apiKey']), self.autosolve_constants.STATUS)
+        try:
+            self.queue.append(
+                {"message": message, "route": self.cancel_send_routing_key, "exchange": self.fanoutExchangeName})
+            send_result = self.send()
+            if send_result is None:
+                self.log("Cancel all requests sent successfully", self.autosolve_constants.SUCCESS)
+        except Exception as e:
+            self.log(e, self.autosolve_constants.WARNING)
+            self.executor.submit(self.resend, message, self.cancel_send_routing_key, self.fanoutExchangeName)
 
     def send(self):
         req = None
+        message = {}
         try:
             req = self.queue.pop()
             if req is not None:
@@ -254,61 +290,21 @@ class InternalAutoSolve:
                 byte_string = message_string.encode()
                 if self.fanoutExchangeName == exchange:
                     return self.cancel_channel.basic_publish(exchange=exchange,
-                                                    routing_key=route,
-                                                    body=byte_string)
+                                                             routing_key=route,
+                                                             body=byte_string)
 
                 return self.channel.basic_publish(exchange=exchange,
                                                   routing_key=route,
                                                   body=byte_string)
         except Exception as e:
-            print(e)
+            self.log("Error sending message " + str(message["taskId"]), self.autosolve_constants.ERROR)
             if req is not None:
                 self.queue.append(req)
-
-    def send_token_request(self, message):
-        self.log("Sending request message for task: " + str(message['taskId']), self.autosolve_constants.STATUS)
-        try:
-            self.queue.append({"message": message, "route": self.token_send_routing_key, "exchange": self.directExchangeName})
-            send_result = self.connection.add_callback_threadsafe(self.send)
-            if send_result is None:
-                self.log("Message with TaskId: " + str(message['taskId']) + " sent successfully",
-                         self.autosolve_constants.SUCCESS)
-        except Exception as e:
-            self.log(e, self.autosolve_constants.WARNING)
-            self.resend(message, self.token_send_routing_key, self.directExchangeName)
-
-    def cancel_token_request(self, taskId):
-        message = dict()
-        message["taskId"] = taskId
-        message["responseRequired"] = self.should_alert_on_cancel
-
-        self.log("Sending cancel message for task: " + message['taskId'], self.autosolve_constants.STATUS)
-        try:
-            self.queue.append({"message": message, "route": self.cancel_send_routing_key, "exchange": self.fanoutExchangeName})
-            send_result = self.connection.add_callback_threadsafe(self.send)
-            if send_result is None:
-                self.log("Cancel message with TaskId: " + message['taskId'] + " sent successfully",
-                         self.autosolve_constants.SUCCESS)
-        except Exception as e:
-            self.log(e, self.autosolve_constants.WARNING)
-            self.resend(message, self.cancel_send_routing_key, self.fanoutExchangeName)
-
-    def cancel_all_requests(self):
-        message = {"apiKey": self.api_key}
-        self.log("Sending cancel all request for api key ::  " + message['apiKey'], self.autosolve_constants.STATUS)
-        try:
-            self.queue.append(
-                {"message": message, "route": self.cancel_send_routing_key, "exchange": self.fanoutExchangeName})
-            send_result = self.connection.add_callback_threadsafe(self.send)
-            if send_result is None:
-                self.log("Cancel all requests sent successfully", self.autosolve_constants.SUCCESS)
-        except Exception as e:
-            self.log(e, self.autosolve_constants.WARNING)
-            self.resend(message, self.cancel_send_routing_key, self.fanoutExchangeName)
+            return False
 
     def resend(self, message, route, exchange):
         time.sleep(5)
-        result = self.connection.add_callback_threadsafe(self.send)
+        result = self.send()
         if result is None:
             self.log("Resend attempt successful", self.autosolve_constants.SUCCESS)
         else:
@@ -321,11 +317,13 @@ class InternalAutoSolve:
 
     def attempt_message_backlog_send(self):
         self.log("Sending messages from backlog", self.autosolve_constants.STATUS)
+        time.sleep(5)
         for message in self.queue:
             self.log("Backlog message sending: " + str(message["message"]["taskId"]), self.autosolve_constants.STATUS)
-            result = self.connection.add_callback_threadsafe(self.send)
+            result = self.send()
             if result is not None:
-                self.log("Backlog resend unsuccessful for message: " + str(message["message"]["taskId"]), self.autosolve_constants.STATUS)
+                self.log("Backlog resend unsuccessful for message: " + str(message["message"]["taskId"]),
+                         self.autosolve_constants.STATUS)
 
     ## VALIDATION FUNCTIONS ##
 
@@ -357,9 +355,13 @@ class InternalAutoSolve:
 
     def validate_connection(self):
         self.log("Validating input values with AutoSolve API", self.autosolve_constants.STATUS)
-        url = "https://dash.autosolve.io/rest/" + self.access_token + "/verify/" + self.api_key + "?clientId=" + self.client_key
-        response = requests.get(url)
-        return response.status_code
+        url = "https://dash.autosolve.aycd.io/rest/" + self.access_token + "/verify/" + self.api_key + "?clientId=" + self.client_key
+        try:
+            response = requests.get(url, timeout=5)
+            return response.status_code
+        except Exception as ex:
+            self.log("Validate failed. Reattempting connection", self.autosolve_constants.ERROR)
+            self.handle_connection_error(None, None)
 
     def build_prefix_with_credentials(self, prefix):
         return prefix + "." + self.account_id + "." + self.routing_key
@@ -402,5 +404,3 @@ class InputValueException(Exception):
 
     def __str__(self):
         return 'InputValueException: {}'.format(self.value)
-
-
